// SPDX-License-Identifier: MIT

use anyhow::{anyhow, Context, Result};
use askama::Template;
use axum::{
    body::Bytes,
    extract::{
        ws::{self, WebSocket, WebSocketUpgrade},
        Extension,
    },
    middleware,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Router,
};
use axum_typed_multipart::{FieldData, TryFromMultipart, TypedMultipart};
use futures::{sink::SinkExt, stream::StreamExt};
use once_cell::sync::OnceCell;
use parking_lot::Mutex;
use std::collections::HashMap;
use std::path::Path;
use tokio::fs;
use tokio::io::AsyncWriteExt;
use tower_sessions::Session;

use crate::common::{
    self, check_auth, get_title, html_template::HtmlTemplate, stream_command, CommandOpts,
    LoggedIn, Title,
};
use crate::error::{ErrorString, ErrorStringResult, PageResult};

static VPN_WS: OnceCell<Mutex<Option<WebSocket>>> = OnceCell::new();

#[derive(TryFromMultipart)]
struct VpnParam {
    setting_name: String,
    conf_file: FieldData<Bytes>,
    auth_type: String,
    username: String,
    password: String,
    cert_file: FieldData<Bytes>,
    key_file: FieldData<Bytes>,
    key_pass: String,
}

#[derive(Template)]
#[template(path = "../src/vpn/templates/vpn.html")]
struct VpnTemplate {
    setting_name: String,
    vpn_addr: HashMap<&'static str, String>,
}

pub fn routes() -> Router {
    VPN_WS
        .set(Mutex::new(None))
        .expect("VPN_WS once cell already set!");

    Router::new()
        .route("/vpn", get(vpn))
        .route("/vpn_setup", post(vpn_setup))
        .route("/vpn_delete", post(vpn_delete))
        .route("/vpn_connect", post(vpn_connect))
        .route("/vpn_disconnect", post(vpn_disconnect))
        .route("/vpn_ws", get(vpn_ws))
        .route_layer(middleware::from_fn(check_auth))
        .route_layer(middleware::from_fn(|request, next| {
            get_title(request, next, "./vpn")
        }))
}

async fn vpn_ws(session: Session, ws: WebSocketUpgrade) -> impl IntoResponse {
    if !cfg!(debug_assertions) && !session.logged_in().await {
        return Err(ErrorString {
            error: "not logged in".to_string(),
        });
    }
    Ok(ws.on_upgrade(vpn_socket_handler))
}

async fn vpn_socket_handler(socket: WebSocket) {
    *VPN_WS
        .get()
        .expect("VPN_WS once cell not initialized")
        .lock() = Some(socket);
}

async fn vpn(Extension(title): Extension<Title>) -> PageResult {
    let args = &["vpn_info.sh"];
    let output = common::exec_command(args).await?;
    let setting_name = String::from_utf8_lossy(&output.stdout).trim().to_string();

    let interface = "tun0";
    let vpn_addr = common::current_address(interface, interface).await?;

    let template = VpnTemplate {
        setting_name,
        vpn_addr,
    };
    Ok(HtmlTemplate::new(title.0, template).into_response())
}

async fn vpn_setup(TypedMultipart(vpn_param): TypedMultipart<VpnParam>) -> ErrorStringResult {
    // Delete current settings before setup.
    let args = &["vpn_delete.sh"];
    common::exec_command(args).await?;

    let conf_file_contents = vpn_param.conf_file.contents;
    let conf_file_name = vpn_param
        .conf_file
        .metadata
        .file_name
        .context("Config file name is none")?;
    if let Err(e) = save_vpn_file(&conf_file_name, &conf_file_contents).await {
        delete_vpn_dir().await?;
        return Err(anyhow!("Could not save file: {}.", e).into());
    }

    let mut args = vec![
        "vpn_setup.sh".to_string(),
        vpn_param.setting_name,
        vpn_param.auth_type.clone(),
        conf_file_name,
    ];

    if vpn_param.auth_type == "userpass" {
        args.push("--user".to_string());
        args.push(vpn_param.username);
        args.push("--password".to_string());
        args.push(vpn_param.password);
    } else {
        let cert_file_contents = vpn_param.cert_file.contents;
        let cert_file_name = match vpn_param.cert_file.metadata.file_name {
            Some(f) => f,
            None => return Err(anyhow!("Cert file name is none.").into()),
        };
        if !cert_file_contents.is_empty() && !cert_file_name.is_empty() {
            if let Err(e) = save_vpn_file(&cert_file_name, &cert_file_contents).await {
                delete_vpn_dir().await?;
                return Err(anyhow!("Could not save file: {}.", e).into());
            }
            args.push("--cert".to_string());
            args.push(cert_file_name);
        }

        let key_file_contents = vpn_param.key_file.contents;
        let key_file_name = match vpn_param.key_file.metadata.file_name {
            Some(f) => f,
            None => return Err(anyhow!("Key file name is none.").into()),
        };
        if !key_file_contents.is_empty() && !key_file_name.is_empty() {
            if let Err(e) = save_vpn_file(&key_file_name, &key_file_contents).await {
                delete_vpn_dir().await?;
                return Err(anyhow!("Could not save file: {}.", e).into());
            }
            args.push("--key".to_string());
            args.push(key_file_name);
        }
        if !vpn_param.key_pass.is_empty() {
            args.push("--askpass".to_string());
            args.push(vpn_param.key_pass);
        }
    }
    vpn_start(args).await?;
    Ok(().into_response())
}

async fn vpn_start(args: Vec<String>) -> Result<()> {
    let Some(mut socket) = VPN_WS
        .get()
        .expect("VPN_WS once cell not initialized")
        .lock()
        .take()
    else {
        return Err(anyhow!("websocket does not exist."));
    };
    let (mut output, _) = (&mut socket).split();
    if let Err(e) = stream_command(&args, &CommandOpts::default(), &mut output).await {
        let _ = output.send(ws::Message::Close(None)).await;
        return Err(anyhow!("Could not start vpn: {}", e));
    };
    let _ = output.send(ws::Message::Close(None)).await;
    Ok(())
}

async fn vpn_delete() -> PageResult {
    let args = &["vpn_delete.sh"];
    common::exec_command(args).await?;

    Ok(Redirect::to("/vpn").into_response())
}

async fn vpn_connect() -> PageResult {
    let args = &["podman.sh", "start", common::OPENVPN_CONTAINER];
    common::exec_command(args).await?;

    Ok(Redirect::to("/vpn").into_response())
}

async fn vpn_disconnect() -> PageResult {
    let args = &["podman.sh", "stop", common::OPENVPN_CONTAINER];
    common::exec_command(args).await?;

    Ok(Redirect::to("/vpn").into_response())
}

async fn save_vpn_file(file_name: &str, data: &[u8]) -> Result<()> {
    let conf_dir = Path::new(common::OPENVPN_CONF_DIR);
    if !conf_dir.exists() {
        fs::create_dir_all(conf_dir).await?;
    }
    let path = conf_dir.join(file_name);
    let mut file = fs::File::create(path).await?;
    file.write_all(data).await?;
    Ok(())
}

async fn delete_vpn_dir() -> Result<()> {
    fs::remove_dir_all(common::OPENVPN_CONF_DIR).await?;
    Ok(())
}
